
#include <ctype.h>

#include <errno.h>

#include <fcntl.h>

#include <paths.h>

#include <pwd.h>

#include <signal.h>

#include <string.h>

#include <termios.h>

#include <unistd.h>

#include "readpassphrase.h"

#ifndef TCSASOFT

#define TCSASOFT 0

#endif
static volatile sig_atomic_t signo;
static void handler1 (int);

char *readpassphrase (const char *prompt, char *buf, size_t bufsiz, int flags) {
    ssize_t nr;
    int input, output, save_errno;
    char ch, *p, *end;
    struct termios term, oterm;
    struct sigaction sa, savealrm, saveint, savehup, savequit, saveterm;
    struct sigaction savetstp, savettin, savettou, savepipe;
    if (bufsiz == 0) {
        errno = EINVAL;
        return (NULL);
    }
restart :
    signo = 0;
    nr = -1;
    save_errno = 0;
    if ((flags & RPP_STDIN) || (input = output = open (_PATH_TTY, O_RDWR)) == -1) {
        if (flags & RPP_REQUIRE_TTY) {
            errno = ENOTTY;
            return (NULL);
        }
        input = STDIN_FILENO;
        output = STDERR_FILENO;
    }
    sigemptyset (& sa.sa_mask);
    sa.sa_flags = 0;
    sa.sa_handler = handler1;
    (void) sigaction (SIGALRM, &sa, &savealrm);
    (void) sigaction (SIGHUP, &sa, &savehup);
    (void) sigaction (SIGINT, &sa, &saveint);
    (void) sigaction (SIGPIPE, &sa, &savepipe);
    (void) sigaction (SIGQUIT, &sa, &savequit);
    (void) sigaction (SIGTERM, &sa, &saveterm);
    (void) sigaction (SIGTSTP, &sa, &savetstp);
    (void) sigaction (SIGTTIN, &sa, &savettin);
    (void) sigaction (SIGTTOU, &sa, &savettou);
    if (input != STDIN_FILENO && tcgetattr (input, &oterm) == 0) {
        memcpy (& term, & oterm, sizeof (term));
        if (!(flags & RPP_ECHO_ON))
            term.c_lflag &= ~(ECHO | ECHONL);

        #ifdef VSTATUS
        if (term.c_cc[VSTATUS] != _POSIX_VDISABLE)
            term.c_cc[VSTATUS] = _POSIX_VDISABLE;

        #endif
        (void) tcsetattr (input, TCSAFLUSH | TCSASOFT, &term);
    }
    else {
        memset (& term, 0, sizeof (term));
        term.c_lflag |= ECHO;
        memset (& oterm, 0, sizeof (oterm));
        oterm.c_lflag |= ECHO;
    }
    if (signo != SIGTTOU && signo != SIGTTIN) {
        if (!(flags & RPP_STDIN))
            (void) write (output, prompt, strlen (prompt));
        end = buf + bufsiz - 1;
        p = buf;
        while ((nr = read (input, &ch, 1)) == 1 && ch != '\n' && ch != '\r') {
            if (p < end) {
                if ((flags & RPP_SEVENBIT))
                    ch &= 0x7f;
                if (isalpha (ch)) {
                    if ((flags & RPP_FORCELOWER))
                        ch = (char) tolower (ch);
                    if ((flags & RPP_FORCEUPPER))
                        ch = (char) toupper (ch);
                }
                *p++ = ch;
            }
        }
        *p = '\0';
        save_errno = errno;
        if (!(term.c_lflag & ECHO))
            (void) write (output, "\n", 1);
    }
    if (memcmp (&term, &oterm, sizeof (term)) != 0) {
        while (tcsetattr (input, TCSAFLUSH | TCSASOFT, &oterm) == -1 && errno == EINTR)
            continue;
    }
    (void) sigaction (SIGALRM, &savealrm, NULL);
    (void) sigaction (SIGHUP, &savehup, NULL);
    (void) sigaction (SIGINT, &saveint, NULL);
    (void) sigaction (SIGQUIT, &savequit, NULL);
    (void) sigaction (SIGPIPE, &savepipe, NULL);
    (void) sigaction (SIGTERM, &saveterm, NULL);
    (void) sigaction (SIGTSTP, &savetstp, NULL);
    (void) sigaction (SIGTTIN, &savettin, NULL);
    (void) sigaction (SIGTTOU, &savettou, NULL);
    if (input != STDIN_FILENO)
        (void) close (input);
    if (signo) {
        kill (getpid (), signo);
        switch (signo) {
        case SIGTSTP :
        case SIGTTIN :
        case SIGTTOU :
            goto restart;
        }
    }
    if (save_errno)
        errno = save_errno;
    return (nr == -1 ? NULL : buf);
}

void handler1 (int s) {
    signo = s;
}

